#!/usr/bin/env python3

from __future__ import absolute_import, division, print_function
from aro_msgs.srv import GenerateFrontier, PlanPath, PlanPathRequest, PlanPathResponse
from geometry_msgs.msg import Pose, Pose2D, PoseStamped, PoseArray, Quaternion, Transform, TransformStamped
from aro_msgs.msg import FollowPathAction, FollowPathFeedback, FollowPathResult, FollowPathGoal
from aro_msgs.msg import Path
from std_srvs.srv import SetBool, SetBoolRequest
import actionlib
import numpy as np
from ros_numpy import msgify, numpify
import rospy
from tf.transformations import euler_from_quaternion, quaternion_from_euler
from tf2_py import TransformException
import tf2_ros
from threading import RLock

np.set_printoptions(precision=3)


def slots(msg):
    """Return message attributes (slots) as list."""
    return [getattr(msg, var) for var in msg.__slots__]


def array(msg):
    """Return message attributes (slots) as array."""
    return np.array(slots(msg))


def pose2to3(pose2):
    """Convert Pose2D to Pose."""
    pose3 = Pose()
    pose3.position.x = pose2.x
    pose3.position.y = pose2.y
    rpy = 0.0, 0.0, pose2.theta
    q = quaternion_from_euler(*rpy)
    pose3.orientation = Quaternion(*q)
    return pose3


def tf3to2(tf):
    """Convert Transform to Pose2D."""
    pose2 = Pose2D()
    pose2.x = tf.translation.x
    pose2.y = tf.translation.y
    rpy = euler_from_quaternion(slots(tf.rotation))
    pose2.theta = rpy[2]
    return pose2


class Explorer(object):
    def __init__(self):
        rospy.loginfo('Initializing explorer.')
        self.map_frame = rospy.get_param('~map_frame', 'map')
        self.odom_frame = rospy.get_param('~odom_frame', 'odom')
        self.robot_frame = rospy.get_param('~robot_frame', 'base_footprint')
        self.plan_interval = rospy.Duration(rospy.get_param('~plan_interval', 10.0))
        self.goal_reached_dist = rospy.get_param('~goal_reached_dist', 0.2)
        self.retries = rospy.get_param('~retries', 3)
        self.run_mode = rospy.get_param('~run_mode', 'eval')
        assert self.retries >= 1

        # Mind thread safety which is not guaranteed in rospy when multiple callbacks access the same data.
        self.lock = RLock()

        # Using exploration / frontier.py
        rospy.wait_for_service('get_closest_frontier')
        self.get_closest_frontier = rospy.ServiceProxy('get_closest_frontier', GenerateFrontier)
        rospy.wait_for_service('get_random_frontier')
        self.get_random_frontier = rospy.ServiceProxy('get_random_frontier', GenerateFrontier)
        rospy.wait_for_service('get_best_value_frontier')
        
        # Utilize stop_simulation service when submitting for faster evaluation. Refer to evaluathor package for details.

        #rospy.wait_for_service('stop_simulation')
        #self.stop_sim = rospy.ServiceProxy('stop_simulation', SetBool)

        self.tf = tf2_ros.Buffer()
        self.tf_sub = tf2_ros.TransformListener(self.tf)
        rospy.loginfo('Initializing services and publishers.')

        rospy.wait_for_service('plan_path')
        self.plan_path = rospy.ServiceProxy('plan_path', PlanPath)

        # For exploration / path_follower
        
        self.path_pub = rospy.Publisher('path', Path, queue_size=2)
        
        self.poses_pub = rospy.Publisher('poses', PoseArray, queue_size=2)

        self.action_client = actionlib.SimpleActionClient('follow_path', FollowPathAction)

        self.tf = tf2_ros.Buffer()
        self.tf_sub = tf2_ros.TransformListener(self.tf)

        # Wait for initial pose.
        self.initPose = None
        while self.initPose == None:
            try:
                self.initPose = self.get_robot_pose()
                rospy.loginfo('Explorer initial pose:')
                rospy.loginfo(self.initPose)
            except:
                rospy.loginfo('Robot odometry is not ready yet.')
                self.initPose = None
                rospy.sleep(1.0)

        # TODO: Create an exploration control loop state machine using timer and packages created for your homeworks. Good luck.
        rospy.loginfo('Initializing timer.')

    def lookup_transform(self, target_frame, source_frame, time,
                         no_wait_frame=None, timeout=rospy.Duration.from_sec(0.0)):
        if no_wait_frame is None or no_wait_frame == target_frame:
            return self.tf.lookup_transform(target_frame, source_frame, time, timeout=timeout)

        tf_n2t = self.tf.lookup_transform(self.map_frame, self.odom_frame, rospy.Time())
        tf_s2n = self.tf.lookup_transform(self.odom_frame, self.robot_frame, time, timeout=timeout)
        tf_s2t = TransformStamped()
        tf_s2t.header.frame_id = target_frame
        tf_s2t.header.stamp = time
        tf_s2t.child_frame_id = source_frame
        tf_s2t.transform = msgify(Transform,
                                  np.matmul(numpify(tf_n2t.transform),
                                            numpify(tf_s2n.transform)))
        return tf_s2t

    def marker_cb(self, msg):
        self.marker_pose = msg.pose
        rospy.loginfo('Marker localized at pose: [%.1f,%.1f].', self.marker_pose.position.x, self.marker_pose.position.y)

    def get_robot_pose(self):
        tf = self.lookup_transform(self.map_frame, self.robot_frame, rospy.Time.now(),
                                   timeout=rospy.Duration.from_sec(0.5), no_wait_frame=self.odom_frame)
        pose = tf3to2(tf.transform)
        return pose

if __name__ == '__main__':
    rospy.init_node('explorer', log_level=rospy.INFO)
    node = Explorer()
    rospy.spin()
